use std::{
    collections::HashMap,
    ops::{Deref, DerefMut},
};

use anyhow::Result;

use crate::command::*;

use super::DeviceBackend;

#[derive(Debug, Clone, Copy, PartialEq, Eq, PartialOrd, Ord, Hash)]
pub enum DeviceAttributes {
    Power,
    Status,
    Mode,
    Additional,
    Door,
    RinseAid,
    Salt,
    ChildLock,
    UV,
    Dry,
    DryStatus,
    Storage,
    StorageStatus,
    TimeRemaining,
    Progress,
    StorageRemaining,
    Temperature,
    Humidity,
    Waterswitch,
    WaterLack,
    ErrorCode,
    Softwater,
    WrongOperation,
    Bright,
}

impl DeviceAttributes {
    fn as_str(&self) -> &str {
        match self {
            DeviceAttributes::Power => "power",
            DeviceAttributes::Status => "status",
            DeviceAttributes::Mode => "mode",
            DeviceAttributes::Additional => "additional",
            DeviceAttributes::Door => "door",
            DeviceAttributes::RinseAid => "rinse_aid",
            DeviceAttributes::Salt => "salt",
            DeviceAttributes::ChildLock => "child_lock",
            DeviceAttributes::UV => "uv",
            DeviceAttributes::Dry => "dry",
            DeviceAttributes::DryStatus => "dry_status",
            DeviceAttributes::Storage => "storage",
            DeviceAttributes::StorageStatus => "storage_status",
            DeviceAttributes::TimeRemaining => "time_remaining",
            DeviceAttributes::Progress => "progress",
            DeviceAttributes::StorageRemaining => "storage_remaining",
            DeviceAttributes::Temperature => "temperature",
            DeviceAttributes::Humidity => "humidity",
            DeviceAttributes::Waterswitch => "waterswitch",
            DeviceAttributes::WaterLack => "water_lack",
            DeviceAttributes::ErrorCode => "error_code",
            DeviceAttributes::Softwater => "softwater",
            DeviceAttributes::WrongOperation => "wrong_operation",
            DeviceAttributes::Bright => "bright",
        }
    }

    fn from_str(s: &str) -> Self {
        match s {
            "power" => Self::Power,
            "status" => Self::Status,
            "mode" => Self::Mode,
            "additional" => Self::Additional,
            "door" => Self::Door,
            "rinse_aid" => Self::RinseAid,
            "salt" => Self::Salt,
            "child_lock" => Self::ChildLock,
            "uv" => Self::UV,
            "dry" => Self::Dry,
            "dry_status" => Self::DryStatus,
            "storage" => Self::Storage,
            "storage_status" => Self::StorageStatus,
            "time_remaining" => Self::TimeRemaining,
            "progress" => Self::Progress,
            "storage_remaining" => Self::StorageRemaining,
            "temperature" => Self::Temperature,
            "humidity" => Self::Humidity,
            "waterswitch" => Self::Waterswitch,
            "water_lack" => Self::WaterLack,
            "error_code" => Self::ErrorCode,
            "softwater" => Self::Softwater,
            "wrong_operation" => Self::WrongOperation,
            "bright" => Self::Bright,

            _ => panic!(),
        }
    }
}

pub enum AttributeValue {
    String(Option<String>),
    Bool(bool),
    Int(u8),
}

impl AttributeValue {
    pub fn set(&mut self, value: impl Into<Self>) {
        match (self, value.into()) {
            (Self::String(current), Self::String(new)) => *current = new,
            (Self::String(current), Self::Bool(new)) => *current = Some(new.to_string()),
            (Self::String(current), Self::Int(new)) => *current = Some(new.to_string()),
            (Self::Bool(current), Self::Bool(new)) => *current = new,
            (Self::Int(current), Self::Int(new)) => *current = new,

            _ => panic!(),
        }
    }

    pub fn str(&self) -> Option<&str> {
        match self {
            Self::String(opt) => opt.as_ref().map(|s| s.as_str()),
            _ => panic!(),
        }
    }

    pub fn bool(&self) -> bool {
        match self {
            Self::Bool(bool) => *bool,
            _ => panic!(),
        }
    }

    pub fn byte(&self) -> u8 {
        match self {
            Self::Int(b) => *b,
            _ => panic!(),
        }
    }
}

impl From<String> for AttributeValue {
    fn from(value: String) -> Self {
        Self::String(Some(value))
    }
}

impl From<Option<String>> for AttributeValue {
    fn from(value: Option<String>) -> Self {
        Self::String(value)
    }
}

impl From<bool> for AttributeValue {
    fn from(value: bool) -> Self {
        Self::Bool(value)
    }
}

impl From<u8> for AttributeValue {
    fn from(value: u8) -> Self {
        Self::Int(value)
    }
}

pub struct E1 {
    modes: HashMap<u32, String>,
    attributes: HashMap<DeviceAttributes, AttributeValue>,

    status: [&'static str; 5],
    progress: [&'static str; 6],

    device_protocol_version: u8,
}

impl E1 {
    pub fn new() -> Result<Self> {
        let modes = [
            (0x0, "Neutral Gear"),
            (0x1, "Auto"),
            (0x2, "Heavy"),
            (0x3, "Normal"),
            (0x4, "Energy Saving"),
            (0x5, "Delicate"),
            (0x6, "Hour"),
            (0x7, "Quick"),
            (0x8, "Rinse"),
            (0x9, "90min"),
            (0xA, "Self Clean"),
            (0xB, "Fruit Wash"),
            (0xC, "Self Define"),
            (0xD, "Germ"),
            (0xE, "Bowl Wash"),
            (0xF, "Kill Germ"),
            (0x10, "Sea Food Wash"),
            (0x12, "Hot Pot Wash"),
            (0x13, "Quiet"),
            (0x14, "Less Wash"),
            (0x16, "Oil Net Wash"),
            (0x19, "Cloud Wash"),
        ]
        .into_iter()
        .map(|(i, s)| (i, s.to_string()))
        .collect();

        let attributes = [
            (DeviceAttributes::Power, AttributeValue::Bool(false)),
            (DeviceAttributes::Status, AttributeValue::String(None)),
            (DeviceAttributes::Mode, AttributeValue::Int(0)),
            (DeviceAttributes::Additional, AttributeValue::Int(0)),
            (DeviceAttributes::UV, AttributeValue::Bool(false)),
            (DeviceAttributes::Dry, AttributeValue::Bool(false)),
            (DeviceAttributes::DryStatus, AttributeValue::Bool(false)),
            (DeviceAttributes::Door, AttributeValue::Bool(false)),
            (DeviceAttributes::RinseAid, AttributeValue::Bool(false)),
            (DeviceAttributes::Salt, AttributeValue::Bool(false)),
            (DeviceAttributes::ChildLock, AttributeValue::Bool(false)),
            (DeviceAttributes::Storage, AttributeValue::Bool(false)),
            (DeviceAttributes::StorageStatus, AttributeValue::Bool(false)),
            (
                DeviceAttributes::TimeRemaining,
                AttributeValue::String(None),
            ),
            (DeviceAttributes::Progress, AttributeValue::String(None)),
            (
                DeviceAttributes::StorageRemaining,
                AttributeValue::String(None),
            ),
            (DeviceAttributes::Temperature, AttributeValue::String(None)),
            (DeviceAttributes::Humidity, AttributeValue::String(None)),
            (DeviceAttributes::Waterswitch, AttributeValue::Bool(false)),
            (DeviceAttributes::WaterLack, AttributeValue::Bool(false)),
            (DeviceAttributes::ErrorCode, AttributeValue::String(None)),
            (DeviceAttributes::Softwater, AttributeValue::Int(0)),
            (
                DeviceAttributes::WrongOperation,
                AttributeValue::String(None),
            ),
            (DeviceAttributes::Bright, AttributeValue::Int(0)),
        ]
        .into_iter()
        .collect();

        let status = ["Off", "Idle", "Delay", "Running", "Error"];
        let progress = ["Idle", "Pre-wash", "Wash", "Rinse", "Dry", "Complete"];

        Ok(Self {
            modes,
            attributes,
            status,
            progress,

            device_protocol_version: 0,
        })
    }
}

impl DeviceBackend for E1 {
    fn build_query(&self) -> CommandRequest {
        CommandQuery::new(self.device_protocol_version).request()
    }

    fn process_message(&self, msg: &[u8]) -> Result<Vec<u8>> {
        let cmd = CommandE1Response::new(msg);

        todo!()
    }

    fn set_attribute(&self, attribute: &str, value: &str) -> () {
        match DeviceAttributes::from_str(attribute) {
            DeviceAttributes::Power => (),
            DeviceAttributes::ChildLock => (),
            DeviceAttributes::Storage => (),

            _ => (),
        }
    }
}

macro_rules! concat {
    ($vec1:expr, $($vec2:expr,)+) => {{
        $(
            $vec1.extend(&$vec2);
        )+

        $vec1
    }};
}

pub struct CommandE1Base {
    command: CommandRequest,
}

impl CommandE1Base {
    pub fn new(device_protocol_version: u8, message_type: MessageType, body: Body) -> Self {
        Self {
            command: CommandRequest::new(device_protocol_version, 0xE1, message_type, body),
        }
    }
}

impl Deref for CommandE1Base {
    type Target = CommandRequest;

    fn deref(&self) -> &Self::Target {
        &self.command
    }
}

impl DerefMut for CommandE1Base {
    fn deref_mut(&mut self) -> &mut Self::Target {
        &mut self.command
    }
}

pub struct CommandPower {
    command: CommandE1Base,
}

impl CommandPower {
    pub fn new(device_protocol_version: u8) -> Self {
        Self {
            command: CommandE1Base::new(
                device_protocol_version,
                MessageType::Set,
                Body::from(([0x00; 4].as_slice(), 0x08)),
            ),
        }
    }

    pub fn power(&self) -> bool {
        self.body()[1] == 0x01
    }

    pub fn set_power(&mut self, power: bool) {
        self.command.body_mut()[1] = if power { 0x01 } else { 0x00 };
    }

    pub fn request(self) -> CommandRequest {
        self.command.command
    }
}

impl Deref for CommandPower {
    type Target = CommandRequest;

    fn deref(&self) -> &Self::Target {
        &self.command
    }
}

pub struct CommandLock {
    command: CommandE1Base,
}

impl CommandLock {
    pub fn new(device_protocol_version: u8) -> Self {
        Self {
            command: CommandE1Base::new(
                device_protocol_version,
                MessageType::Set,
                Body::from((concat!(vec![0x04], vec![0x00; 36],).as_slice(), 0x08)),
            ),
        }
    }

    pub fn lock(&self) -> bool {
        self.body()[1] == 0x03
    }

    pub fn set_lock(&mut self, lock: bool) {
        self.command.body_mut()[1] = if lock { 0x03 } else { 0x04 }
    }

    pub fn request(self) -> CommandRequest {
        self.command.command
    }
}

impl Deref for CommandLock {
    type Target = CommandRequest;

    fn deref(&self) -> &Self::Target {
        &self.command
    }
}

pub struct CommandStorage {
    command: CommandE1Base,
}

impl CommandStorage {
    pub fn new(device_protocol_version: u8) -> Self {
        Self {
            command: CommandE1Base::new(
                device_protocol_version,
                MessageType::Set,
                Body::from((
                    concat!(vec![0x00; 4], vec![0xFF; 6], vec![0x00; 27],).as_slice(),
                    0x81,
                )),
            ),
        }
    }

    pub fn storage(&self) -> bool {
        self.body()[4] == 0x01
    }

    pub fn set_storage(&mut self, storage: bool) {
        self.command.body_mut()[4] = if storage { 0x01 } else { 0x00 }
    }

    pub fn request(self) -> CommandRequest {
        self.command.command
    }
}

impl Deref for CommandStorage {
    type Target = CommandRequest;

    fn deref(&self) -> &Self::Target {
        &self.command
    }
}

pub struct CommandQuery {
    command: CommandE1Base,
}

impl CommandQuery {
    pub fn new(device_protocol_version: u8) -> Self {
        Self {
            command: CommandE1Base::new(
                device_protocol_version,
                MessageType::Query,
                Body::from(([].as_slice(), 0x00)),
            ),
        }
    }

    pub fn request(self) -> CommandRequest {
        self.command.command
    }
}

pub struct CommandE1Response {
    command: CommandResponse,
    start: bool,
}

impl CommandE1Response {
    pub fn new(msg: &[u8]) -> Self {
        let command = CommandResponse::new(msg);

        Self {
            command,
            start: false,
        }
    }

    pub fn update_attributes(
        &mut self,
        attributes: &mut HashMap<DeviceAttributes, AttributeValue>,
    ) {
        let message_type = self.command.header().message_type();
        let body = self.command.body();
        let body_type = self.command.body().body_type();

        if (message_type == MessageType::Set && 0 <= body_type && body_type <= 7)
            || ((message_type == MessageType::Query || message_type == MessageType::Notify1)
                && body_type == 0)
        {
            attributes
                .get_mut(&DeviceAttributes::Power)
                .unwrap()
                .set(body[1] > 0);
            attributes
                .get_mut(&DeviceAttributes::Status)
                .unwrap()
                .set(body[1]);
            attributes
                .get_mut(&DeviceAttributes::Mode)
                .unwrap()
                .set(body[2]);
            attributes
                .get_mut(&DeviceAttributes::Additional)
                .unwrap()
                .set(body[3]);

            // 0 - open, 1 - close
            attributes
                .get_mut(&DeviceAttributes::Door)
                .unwrap()
                .set((body[5] & 0x01) == 0);

            // 0 - enough, 1 - shortage
            attributes
                .get_mut(&DeviceAttributes::RinseAid)
                .unwrap()
                .set((body[5] & 0x02) > 0);

            // e - enough, 1 - shortage
            attributes
                .get_mut(&DeviceAttributes::Salt)
                .unwrap()
                .set((body[5] & 0x04) > 0);

            let start_pause = (body[5] & 0x08) > 0;

            if start_pause {
                self.start = true;
            } else if attributes[&DeviceAttributes::Status].byte() == 2
                || attributes[&DeviceAttributes::Status].byte() == 3
            {
                self.start = false;
            }

            attributes
                .get_mut(&DeviceAttributes::ChildLock)
                .unwrap()
                .set((body[5] & 0x10) > 0);
            attributes
                .get_mut(&DeviceAttributes::UV)
                .unwrap()
                .set((body[4] & 0x02) > 0);
            attributes
                .get_mut(&DeviceAttributes::Dry)
                .unwrap()
                .set((body[4] & 0x10) > 0);
            attributes
                .get_mut(&DeviceAttributes::DryStatus)
                .unwrap()
                .set((body[4] & 0x20) > 0);
            attributes
                .get_mut(&DeviceAttributes::Status)
                .unwrap()
                .set((body[5] & 0x20) > 0);
            attributes
                .get_mut(&DeviceAttributes::StorageStatus)
                .unwrap()
                .set((body[5] & 0x40) > 0);
            attributes
                .get_mut(&DeviceAttributes::TimeRemaining)
                .unwrap()
                .set(body[6]);
            attributes
                .get_mut(&DeviceAttributes::Progress)
                .unwrap()
                .set(body[9]);
            attributes
                .get_mut(&DeviceAttributes::StorageRemaining)
                .unwrap()
                .set(if body.len() > 18 {
                    body[11]
                } else {
                    false.into()
                });
            attributes
                .get_mut(&DeviceAttributes::Temperature)
                .unwrap()
                .set(body[11]);
            attributes
                .get_mut(&DeviceAttributes::Humidity)
                .unwrap()
                .set(if body.len() > 33 {
                    body[33].into()
                } else {
                    AttributeValue::from(None)
                });
            attributes
                .get_mut(&DeviceAttributes::Waterswitch)
                .unwrap()
                .set((body[4] & 0x04) > 0);
            attributes
                .get_mut(&DeviceAttributes::WaterLack)
                .unwrap()
                .set((body[5] & 0x80) > 0);
            attributes
                .get_mut(&DeviceAttributes::ErrorCode)
                .unwrap()
                .set(body[10]);
            attributes
                .get_mut(&DeviceAttributes::Softwater)
                .unwrap()
                .set(body[13]);
            attributes
                .get_mut(&DeviceAttributes::WrongOperation)
                .unwrap()
                .set(body[16]);
            attributes
                .get_mut(&DeviceAttributes::Bright)
                .unwrap()
                .set(if body.len() > 24 {
                    body[24].into()
                } else {
                    AttributeValue::from(None)
                });
        }
    }
}
